@@ -1,5 +1,6 @@
-from .logs.core import configure_logger
-from pyld import jsonld
+import pyld
+
+from dexa_sdk.logs.core import configure_logger
 
 # Configure loguru logger
 configure_logger()
@@ -10,7 +11,7 @@ cache = {}
 
 def caching_document_loader(url, options):
     """Simple in-memory cache for JSONLD context resolutions"""
-    loader = jsonld.requests_document_loader()
+    loader = pyld.jsonld.requests_document_loader()
     if url in cache:
         return cache[url]
     resp = loader(url)
@@ -18,4 +19,4 @@ def caching_document_loader(url, options):
     return resp
 
 
-jsonld.set_document_loader(caching_document_loader)
+pyld.jsonld.set_document_loader(caching_document_loader)
